# Author: Cameron F. Abrams, <cfa22@drexel.edu>

from .eos import CubicEOS, sqrt_2
from dataclasses import dataclass
import numpy as np
import logging

logger = logging.getLogger(__name__)

@dataclass
class PengRobinsonEOS(CubicEOS):
    """
    Pure-component Peng-Robinson equation of state
    """

    description: str = "Peng-Robinson Equation of State"

    @property
    def kappa(self):
        """
        kappa parameter for Peng-Robinson EOS
        """
        return 0.37464 + 1.54226 * self.omega - 0.26992 * self.omega**2
    
    @property
    def alpha(self):
        """
        alpha parameter for Peng-Robinson EOS
        """
        return (1 + self.kappa * (1 - np.sqrt(self.T / self.Tc)))**2

    def _calc_a(self):
        """
        Calculates parameter a for Peng-Robinson EOS
        """
        return 0.45724 * self.R**2 * self.Tc**2 / self.Pc * self.alpha
    
    def _calc_b(self):
        """
        Calculates parameter b for Peng-Robinson EOS
        """
        return 0.07780 * self.R * self.Tc / self.Pc

    def _calc_da_dT(self):
        """
        Temperature derivative of parameter a for Peng-Robinson EOS
        """
        return -self.a * self.kappa / np.sqrt(self.alpha * self.T * self.Tc)
    
    @property
    def cubic_coeff(self):
        """
        cubic coefficients for Peng-Robinson EOS
        """
        A, B = self.A, self.B
        return np.array([1.0, -1 + B, A - 3 * B**2 - 2*B, -A * B + B**2 + B**3])

    @property
    def lrfrac(self) -> np.ndarray:
        """
        Helper property for logarithmic fraction term in Peng-Robinson EOS
        """
        z = self.Z
        B = self.B
        num_arg = z + (1 + sqrt_2) * B
        den_arg = z + (1 - sqrt_2) * B
        return np.log(num_arg / den_arg)
        
    def _calc_h_departure(self) -> np.ndarray:
        """
        Enthalpy departure for Peng-Robinson EOS
        """
        z = self.Z
        return self.R * self.T * (z - 1) + (self.T * self.da_dT - self.a) / (2 * sqrt_2 * self.b) * self.lrfrac

    def _calc_s_departure(self) -> np.ndarray:
        """
        Entropy departure for Peng-Robinson EOS
        """
        z = self.Z
        return self.R * np.log(z - self.B) + self.da_dT/(2 * sqrt_2 * self.b) * self.lrfrac
    
    def _calc_logphi(self) -> np.ndarray:
        """
        natural log of fugacity coefficient for Peng-Robinson EOS
        """
        z = self.Z
        return z - 1 - np.log(z - self.B) - self.A / (2 * np.sqrt(2) * self.B) * self.lrfrac




# Author: Cameron F. Abrams, <cfa22@drexel.edu>
"""
Pure-component cubic equations of state and their state calculations
"""
from __future__ import annotations
import logging
import numpy as np

from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass, field
from scipy.optimize import root_scalar

from sandlermisc.gas_constant import GasConstant
from sandlermisc.thermals import DeltaH_IG, DeltaS_IG
from sandlerprops.compound import Compound
from sandlerprops.properties import get_database

logger = logging.getLogger(__name__)

sqrt_2 = np.sqrt(2)

@dataclass
class CubicEOS(ABC):
    """
    Abstract class for all Cubic equations of state.
    """
    _cache: dict = field(default_factory=dict, init=False, repr=False)
    """ Cache for computed properties """
    _input_state: dict | None = field(default=None, init=False, repr=False)
    """ Snapshot of input field values for cache validation """

    _INPUT_FIELDS = frozenset(['T', 'P', 'x', 'Tc', 'Pc', 'omega', 'Cp', 'Tref', 'Pref_MPa', 'phase', 'pressure_unit', 'volume_unit'])
    """ Fields that affect calculations and cache validity """

    def __setattr__(self, name, value):
        """ Clear cache on input field changes """
        if name in self._INPUT_FIELDS and hasattr(self, '_cache'):
            self._cache.clear()
        super().__setattr__(name, value)

    def _get_current_input_state(self):
        """ Snapshot of current input field values. """
        return {field: getattr(self, field) for field in self._INPUT_FIELDS}

    pressure_unit: str = 'MPa' # MPa
    """ units of pressure, lower-case """
    volume_unit: str = 'm3'
    """ units of volume, lower-case """

    phase: str = 'unspecified' 
    """ strict phase requirement flag; 'vapor', 'liquid', 'unspecified' (multiple roots) """

    P: float = None
    """ state pressure """
    T: float = None
    """ state temperature """
    x: float = None
    """ vapor fraction in two-phase region; None if single phase """

    Pc: float = None
    """ critical pressure """
    Tc: float = None
    """ critical temperature """
    omega: float = None
    """ acentricity factor, dimensionless """

    Cp: float | list[float] | dict [str, float] | None = None
    """ heat capacity data for ideal-gas contributions """
    
    Tref: float = 298.15
    """ reference temperature for 'absolute' internal energy, enthalpy, and entropy calculations """
    Pref_MPa: float = 0.1
    """ reference pressure for 'absolute' entropy calculations, in MPa """

    logiter: bool = False
    """ flag for logging iterations in phase calculations """
    maxiter: int = 100
    """ maximum iterations in phase calculations """
    epsilon: float = 1.e-5
    """ fugacity tolerance in phase calculations (Fig. 7.5-1 in Sandler 5th ed) """

    def is_cache_stale(self, property_name: str = None):
        """
        Check if cache is stale.
        
        Parameters
        ----------
        property_name: str
            Specific property to check, or None for general staleness
        
        Returns
        -------
        bool or dict: True/False if property_name given, else dict of changed inputs
        """
        if property_name:
            return property_name not in self._cache
        
        # Show which inputs changed
        if self._input_state is None:
            return self._INPUT_FIELDS  # All inputs "changed" (initial state)
        
        current = self._get_current_input_state()
        changed = {k for k in self._INPUT_FIELDS if current[k] != self._input_state.get(k)}
        return changed if changed else False

    def clone(self) -> CubicEOS:
        """ Return a copy of the current EOS object """
        return type(self)(
            pressure_unit=self.pressure_unit,
            volume_unit=self.volume_unit,
            phase=self.phase,
            P=self.P,
            T=self.T,
            Pc=self.Pc,
            Tc=self.Tc,
            omega=self.omega,
            logiter=self.logiter,
            maxiter=self.maxiter,
            epsilon=self.epsilon,
            Tref=self.Tref,
            Pref_MPa=self.Pref_MPa,
            Cp=deepcopy(self.Cp)
        )

    @property
    def Pv(self):
        """ product of pressure and molar volume """
        return self.P * self.v
    
    @property
    def R_pv(self):
        return GasConstant(self.pressure_unit, self.volume_unit)
    
    @property
    def R(self):
        """ default gas constant value """
        return GasConstant() # J/(mol-K)

    @abstractmethod
    def _calc_a(self):
        pass

    @property
    def a(self):
        """ van der Waals a parameter """
        if not 'a' in self._cache: # cache was cleared on an input change
            self._cache['a'] = self._calc_a()
            self._input_state = self._get_current_input_state()  # Save state
        return self._cache['a']
    
    def _calc_da_dT(self):
        return 0.0

    @property
    def da_dT(self):
        """ derivative of vdW a parameter wrt temperature """
        if not 'da_dT' in self._cache:
            self._cache['da_dT'] = self._calc_da_dT()
            self._input_state = self._get_current_input_state()  # Save state
        return self._cache['da_dT']

    @abstractmethod
    def _calc_b(self):
        pass

    @property
    def b(self):
        """ van der Waals b parameter """
        if not 'b' in self._cache:
            self._cache['b'] = self._calc_b()
            self._input_state = self._get_current_input_state()  # Save state
        return self._cache['b']
    
    @property
    def A(self):
        """ dimensionless vdW a parameter aP/(R^2 T^2) """
        if not 'A' in self._cache:
            self._cache['A'] = self.a * self.P / (self.R * self.T)**2
            self._input_state = self._get_current_input_state()  # Save state
        return self._cache['A']

    @property
    def B(self):
        """ dimensionless vdW b parameter bP/(RT) """
        if not 'B' in self._cache:
            self._cache['B'] = self.b * self.P / (self.R * self.T)
            self._input_state = self._get_current_input_state()  # Save state
        return self._cache['B']

    @property
    @abstractmethod
    def cubic_coeff(self):
        """ coefficients of cubic equation for compressibility factor Z """
        pass

    @property
    def Z(self) -> np.ndarray:
        """
        Computes or retrieves compressibility factor(s) at current state T and P.

        Returns
        -------
        np.ndarray
            Compressibility factor(s) at current state T and P
        """
        if 'Z' not in self._cache:
            self._cache['Z'] = self._solve_for_Z()
            self._input_state = self._get_current_input_state()  # Save state
        return self._cache['Z']

    def _solve_for_Z(self) -> np.ndarray:
        """
        Solve the cubic equation for compressibility factor(s) Z if not cached.

        Returns
        -------
        np.ndarray
            Real roots of the cubic equation representing compressibility factors.
        """
        complx_roots = np.roots(self.cubic_coeff)
        real_roots_idx = np.where(complx_roots.imag==0)[0]
        real_roots = complx_roots[real_roots_idx].real
        if len(real_roots) == 1:
            Z = np.array([real_roots[0]])
        else:
            if self.phase == 'vapor':
                Z = np.array([np.max(real_roots)])
            elif self.phase == 'liquid':
                Z = np.array([np.min(real_roots)])
            else: # unspecified/multiple
                Z = np.array([real_roots[0], real_roots[2]]) # sorted by numpy.roots
        return Z

    @property
    def v(self) -> np.ndarray:
        """
        Computes or retrieves molar volume(s) at current state T and P.
        
        Returns
        -------
        np.ndarray
            Molar volume(s) at current state T and P
        """
        if not 'v' in self._cache:
            self._cache['v'] = self.Z * self.R_pv * self.T / self.P
            self._input_state = self._get_current_input_state()  # Save state
        return self._cache['v']

    @abstractmethod
    def _calc_h_departure(self) -> np.ndarray:
        """
        Calculate the enthalpy departure at state T and P.
        """
        pass

    @property
    def h_departure(self) -> np.ndarray:
        """
        Computes or retrieves enthalpy departure at current state T and P.

        Returns
        -------
        np.ndarray
            Enthalpy departure at current state T and P
        """
        if not 'h_departure' in self._cache:
            self._cache['h_departure'] = self._calc_h_departure()
            self._input_state = self._get_current_input_state()  # Save state
        return self._cache['h_departure']

    @abstractmethod
    def _calc_s_departure(self) -> np.ndarray:
        """
        Calculate the entropy departure at state T and P.
        """
        pass

    @property
    def s_departure(self) -> np.ndarray:
        """
        Computes or retrieves entropy departure at current state T and P.
        
        Returns
        -------
        np.ndarray
            Entropy departure at current state T and P
        """
        if not 's_departure' in self._cache:
            self._cache['s_departure'] = self._calc_s_departure()
            self._input_state = self._get_current_input_state()  # Save state
        return self._cache['s_departure']
    
    @abstractmethod
    def _calc_logphi(self) -> np.ndarray:
        """
        Calculate the natural log of fugacity coefficient at state T and P.
        """
        pass

    @property
    def logphi(self) -> np.ndarray:
        """
        Computes or retrieves the natural log of fugacity coefficient at current state T and P.

        Returns
        -------
        np.ndarray
            Natural log of fugacity coefficient(s) at current state T and P
        """
        if not 'logphi' in self._cache:
            self._cache['logphi'] = self._calc_logphi()
            self._input_state = self._get_current_input_state()  # Save state
        return self._cache['logphi']
    
    @property
    def phi(self) -> np.ndarray:
        """
        Computes or retrieves fugacity coefficient(s) at current state T and P.
        
        Returns
        -------
        np.ndarray
            Fugacity coefficient(s) at current state T and P
        """
        if not 'phi' in self._cache:
            self._cache['phi'] = np.exp(self.logphi)
        return self._cache['phi']

    @property
    def f(self) -> np.ndarray:
        """
        Computes or retrieves fugacity/fugacities at current state T and P.

        Returns
        -------
        np.ndarray
            Fugacity/fugacities at current state T and P
        """
        if not 'f' in self._cache:
            self._cache['f'] = self.phi * self.P
        return self._cache['f']

    @property
    def Pvap(self) -> float:
        """ 
        Computes or retrieves vapor pressure at current state temperature.  If not calculable, returns np.nan.
        Implements algorithm in Fig. 7.5-1 of Sandler 5th ed.
        """
        if not 'pvap' in self._cache:
            if self.Pc is None or self.Tc is None:
                self._cache['pvap'] = np.nan
                return self._cache['pvap']
            tmp = self.clone()
            tmp.P = tmp.Pc * (tmp.T / tmp.Tc)**8
            pu = self.R._capitalizations.get(self.pressure_unit, self.pressure_unit)
            logger.debug(f'Initial Pvap guess: {tmp.P:.6f} {pu} at T {tmp.T:.6f} (Tc {tmp.Tc:.6f}): f {tmp.f}')
            keepgoing = True
            i = 0
            while keepgoing:
                i += 1
                try:
                    fV, fL = tmp.f
                except:
                    # raise ValueError(f'Pvap calc error: T {tmp.T:.2f}, P {tmp.P:.2f} {pu}, Tc {tmp.Tc:.2f}, Pc {tmp.Pc:.2f} {pu}')
                    self._cache['pvap'] = np.nan
                    return self._cache['pvap']
                err = np.abs(fL / fV - 1)
                if tmp.logiter: logger.debug(f'Pvap: Iter {i}: P {tmp.P:.6f} {pu}, fV {fV:.6f}, fL {fL:.6f}; error {err:.4e}')
                tmp.P *= fL / fV
                if err < tmp.epsilon or i == tmp.maxiter:
                    keepgoing = False
                if i >= tmp.maxiter:
                    logger.warning(f'Reached {i} iterations without convergence; error {err:.4e}')
            self._cache['pvap'] = tmp.P
        return self._cache['pvap']
    
    @property
    def Hvap(self) -> float:
        """ 
        Computes or retrieves enthalpy of vaporization at current state temperature.

        Returns
        -------
        float
            Enthalpy of vaporization at current state temperature. If not calculable, returns np.nan.
        """
        if not 'hvap' in self._cache:
            satd_clone = self.clone()
            satd_clone.P = self.Pvap
            if not np.isnan(satd_clone.P):
                self._cache['hvap'] = satd_clone.h_departure[0] - satd_clone.h_departure[1]
            else:
                self._cache['hvap'] = np.nan
        return self._cache['hvap']
    
    @property
    def Svap(self) -> float:
        """ 
        Computes or retrieves entropy of vaporization at current state temperature.

        Returns
        -------
        float
            Entropy of vaporization at current state temperature.  If not calculable, returns np.nan.
        """
        if not 'svap' in self._cache:
            satd_clone = self.clone()
            satd_clone.P = self.Pvap
            if not np.isnan(satd_clone.P):
                self._cache['svap'] = satd_clone.s_departure[0] - satd_clone.s_departure[1]
            else:
                self._cache['svap'] = np.nan
        return self._cache['svap']

    @property
    def Tsat(self) -> float:
        """ 
        Computes or retrieves saturation temperature at state pressure.
        
        Returns
        -------
        float
            Saturation temperature at state pressure.  If not calculable, returns np.nan.
        """
        if not 'Tsat' in self._cache:
            tmp = self.clone()
            if self.Pc is None or self.Tc is None:
                self._cache['Tsat'] = np.nan
                return self._cache['Tsat']
            tmp.T = tmp.Tc * (tmp.P / tmp.Pc)**0.125
            pu = self.R._capitalizations.get(self.pressure_unit, self.pressure_unit)
            logger.debug(f'Initial Tsat guess: {tmp.T:.6f} at P {tmp.P:.6f} {pu}')
            keepgoing = True
            i = 0
            while keepgoing:
                i += 1
                try:
                    fV, fL = tmp.f
                except:
                    # raise ValueError(f'Error in fugacity calculation in Tsat at T {tmp.T:.2f}, P {tmp.P:.2f} {pu}')
                    self._cache['Tsat'] = np.nan
                    return self._cache['Tsat']
                err = np.abs(fV / fL - 1)
                if tmp.logiter: logger.debug(f'Tsat: Iter {i}: T {tmp.T:.6f}, fV {fV:.6f} {pu}, fL {fL:.6f} {pu}; error {err:.4e}')
                tmp.T *= (fV / fL)**0.125
                if err < tmp.epsilon or i == tmp.maxiter:
                    keepgoing = False
                if i == tmp.maxiter:
                    logger.warning(f'Reached {i} iterations without convergence; error {err:.4e}')
            self._cache['Tsat'] = tmp.T
        return self._cache['Tsat']

    def unit_consistency(self, other: CubicEOS):
        consistent = self.pressure_unit == other.pressure_unit and self.volume_unit == other.volume_unit
        if not consistent:
            raise ValueError('inconsistent units')

    @property
    def h(self) -> np.ndarray:
        """
        Computes absolute enthalpy at state T and P
        
        Returns
        -------
        np.ndarray
            Absolute enthalpy/enthalpies at state T and P
        """
        if self.Cp is None:
            raise ValueError("Cp data required for absolute enthalpy calculation.")
        dH_ideal = DeltaH_IG(self.Tref, self.T, self.Cp)
        logger.debug(f'dh_ideal: {dH_ideal} for Tref {self.Tref} to T {self.T} with Cp {self.Cp}')
        return self.h_departure + dH_ideal

    @property
    def u(self) -> np.ndarray:
        """
        Computes absolute internal energy at state T and P
        
        Returns
        -------
        np.ndarray
            Absolute internal energy/energies at state T and P
        """
        if self.Cp is None:
            raise ValueError("Cp data required for absolute internal energy calculation.")
        # u = h - pv
        return self.h - self.Pv * self.R / self.R_pv
    
    @property
    def Pref_local(self) -> float:
        """ 
        Returns reference pressure in local units

        Returns
        -------
        float
            Reference pressure in local units
        """
        if self.pressure_unit == 'mpa' or self.pressure_unit == 'MPa':
            return self.Pref_MPa
        elif self.pressure_unit == 'bar':
            return self.Pref_MPa * 10.0
        elif self.pressure_unit == 'pa' or self.pressure_unit == 'Pa':
            return self.Pref_MPa * 1.e6
        elif self.pressure_unit == 'kpa' or self.pressure_unit == 'kPa':
            return self.Pref_MPa * 1.e3
        elif self.pressure_unit == 'atm':
            return self.Pref_MPa * 9.86923
        else:
            raise ValueError(f"Unsupported pressure unit: {self.pressure_unit}")

    @property
    def s(self) -> np.ndarray:
        """
        Computes absolute entropy at state T and P
        
        Returns
        -------
        np.ndarray
            Absolute entropy/entropies at state T and P
        """
        if self.Cp is None:
            raise ValueError("Cp data required for absolute entropy calculation.")
        # make sure Pref_local is in correct units (same as self.P)
        dS_ideal = DeltaS_IG(self.Tref, self.Pref_local, self.T, self.P, self.Cp, self.R)
        return self.s_departure + dS_ideal

    def delta_h(self, other: CubicEOS) -> np.ndarray:
        """
        Computes and returns enthalpy change from self to other state
        """
        self.unit_consistency(other)
        return other.h - self.h

    def delta_s(self, other: CubicEOS) -> np.ndarray:
        """
        Computes and returns entropy change from self to other state
        """
        self.unit_consistency(other)
        return other.s - self.s
    
    def delta_pv(self, other: CubicEOS) -> np.ndarray:
        """
        Returns Delta(PV) in thermal (not PV) units 
        """
        self.unit_consistency(other)
        return (other.Pv - self.Pv) * self.R / self.R_pv
    
    def delta_u(self, other: CubicEOS) -> np.ndarray:
        """
        Returns Delta(U) (internal energy)
        """
        return self.delta_h(other) - self.delta_pv(other)
        
    def two_phase_check(self, v: float = None, h: float = None, s: float = None, u: float = None):
        """
        Returns the vapor fraction given one property in the two-phase region.

        Parameters
        ----------
        v: float
            Molar volume
        h: float
            Enthalpy
        s: float
            Entropy
        u: float
            Internal energy

        Returns
        -------
        float
            Vapor fraction, where 0 to 1 signfifies two-phase region, <0 is subcooled liquid, >1 is superheated vapor
        """
        satd_clone = self.clone()
        if satd_clone.T is not None:
            satd_clone.P = self.Pvap
        elif satd_clone.P is not None:
            satd_clone.T = self.Tsat
        if v is not None:
            vV, vL = satd_clone.v
            return (v - vL) / (vV - vL)
        elif h is not None:
            hV, hL = satd_clone.h
            return (h - hL) / (hV - hL)
        elif s is not None:
            sV, sL = satd_clone.s
            return (s - sL) / (sV - sL)
        elif u is not None:
            uV, uL = satd_clone.u
            return (u - uL) / (uV - uL)
        else:
            raise ValueError("One of v, h, s, or u must be provided to determine vapor fraction.")

    def P_solve(self, P_bracket: list[float] = None, v: float = None, h: float = None, s: float = None, u: float = None):
        """
        Solve for pressure given T and one other property (v, h, s, u). Stores result in self.P.

        Parameters
        ----------
        P_bracket: list[float]
            Bracket [P_low, P_high] for root finding
        v: float
            Molar volume
        h: float
            Enthalpy
        s: float
            Entropy
        u: float
            Internal energy
        """
        if v is not None:
            def objective(P):
                self.P = P
                logger.debug(f'Objective eval: P={P}, v={self.v}, target v={v}')
                # pick the root closest to target v
                v_diffs = np.abs(self.v - v)
                v_closest = self.v[np.argmin(v_diffs)]
                return v - v_closest
        elif h is not None:
            def objective(P):
                self.P = P
                logger.debug(f'Objective eval: P={P}, h={self.h}, target h={h}')
                    # pick the root closest to target h
                h_diffs = np.abs(self.h - h)
                h_closest = self.h[np.argmin(h_diffs)]
                return h - h_closest
        elif s is not None:
            def objective(P):
                self.P = P
                logger.debug(f'Objective eval: P={P}, s={self.s}, target s={s}')
                # pick the root closest to target s
                s_diffs = np.abs(self.s - s)
                s_closest = self.s[np.argmin(s_diffs)]
                return s - s_closest
        elif u is not None:
            def objective(P):
                self.P = P
                logger.debug(f'Objective eval: P={P}, u={self.u}, target u={u}')
                # pick the root closest to target u
                u_diffs = np.abs(self.u - u)
                u_closest = self.u[np.argmin(u_diffs)]
                return u - u_closest
        else:
            raise ValueError("One of v, h, s, or u must be provided to solve for pressure.")
        res = root_scalar(objective, bracket=P_bracket)
        self.P = res.root

    def T_solve(self, T_bracket: list[float] = None, v: float = None, h: float = None, s: float = None, u: float = None):
        """
        Solve for temperature given P and one other property (v, h, s, u). Stores result in self.T.

        Parameters
        ----------
        T_bracket: list[float]
            Bracket [T_low, T_high] for root finding
        v: float
            Molar volume
        h: float
            Enthalpy
        s: float
            Entropy
        u: float
            Internal energy
        """
        if v is not None:
            def objective(T):
                self.T = T
                logger.debug(f'Objective eval: T={T}, v={self.v}, target v={v}')
                # pick the root closest to target v
                v_diffs = np.abs(self.v - v)
                v_closest = self.v[np.argmin(v_diffs)]
                return v - v_closest
        elif h is not None:
            def objective(T):
                self.T = T
                logger.debug(f'Objective eval: T={T}, h={self.h}, target h={h}')
                # pick the root closest to target h
                h_diffs = np.abs(self.h - h)
                h_closest = self.h[np.argmin(h_diffs)]
                return h - h_closest
        elif s is not None:
            def objective(T):
                self.T = T
                logger.debug(f'Objective eval: T={T}, s={self.s}, target s={s}')
                # pick the root closest to target s
                s_diffs = np.abs(self.s - s)
                s_closest = self.s[np.argmin(s_diffs)]
                return s - s_closest
        elif u is not None:
            def objective(T):
                self.T = T
                logger.debug(f'Objective eval: T={T}, u={self.u}, target u={u}')
                # pick the root closest to target u
                u_diffs = np.abs(self.u - u)
                u_closest = self.u[np.argmin(u_diffs)]
                return u - u_closest
        else:
            raise ValueError("One of v, h, s, or u must be provided to solve for temperature.")

        res = root_scalar(objective, bracket=T_bracket)
        self.T = res.root

    def solve(self, T: float = None, P: float = None, v: float = None, h: float = None, s: float = None, u: float = None):
        """
        Solve for the missing state variable given two of T, P, v, h, s, u.
        Updates the object's state variables accordingly.

        Parameters
        ----------
        T: float
            Temperature
        P: float
            Pressure
        v: float
            Molar volume
        h: float
            Enthalpy
        s: float
            Entropy
        u: float
            Internal energy
        """
        num_specified = sum(var is not None for var in [T, P, v, h, s, u])
        if num_specified != 2:
            raise ValueError("Exactly two of T, P, v, h, s, or u must be specified.")
        
        self.T = T
        self.P = P

        # Calculate the missing variable using the EOS
        if self.T is not None and self.P is not None:
            pass # all properties can be retrieved now
        elif self.P is None: # T and one other property (not P) is set
            if self.T < self.Tc:
                x = self.two_phase_check(v=v, h=h, s=s, u=u)
                if 0 < x < 1:
                    self.x = x
                    self.P = self.Pvap
                    # done.  All properties can be retrieved now.
                    return
                elif x < 0:
                    # subcooled liquid
                    Plo = self.Pvap
                    Phi = self.Pc
                    phase = 'liquid'
                else:
                    # superheated vapor
                    Phi = self.Pvap
                    Plo = 1.e-5 # guess very low pressure
                    phase = 'vapor'
                self.phase = phase
                self.P_solve([Plo, Phi], v=v, h=h, s=s, u=u)
                self.phase = 'unspecified'
            else: # given T is above Tc, so supercritical
                self.P_solve([1.e-5, self.Pc*5], v=v, h=h, s=s, u=u)
        elif self.T is None:
            if self.P < self.Pc:
                x = self.two_phase_check(v=v, h=h, s=s, u=u)
                logger.debug(f'P {self.P} subcrit: x = {x}')
                if 0 < x < 1:
                    self.x = x
                    self.T = self.Tsat
                    return
                elif x < 0:
                    # subcooled liquid
                    Tlo = 1.0  # guess low temperature
                    Thi = self.Tsat
                    phase = 'liquid'
                else:
                    # superheated vapor
                    Thi = self.Tc*3
                    Tlo = self.Tsat
                    phase = 'vapor'
                self.phase = phase
                self.T_solve([Tlo, Thi], v=v, h=h, s=s, u=u)
                self.phase = 'unspecified'
            else:
                self.T_solve([10.0, self.Tc*3], v=v, h=h, s=s, u=u)
        else:
            raise NotImplementedError("Solving when neither T nor P are set is not implemented yet.")

    def set_compound(self, compound: str | Compound):
        """
        Set critical properties and Cp data from a compound name.

        Parameters
        ----------
        compound: str | Compound
            Name of the compound to retrieve properties for
        """
        db = get_database()
        compound = db.get_compound(compound) if isinstance(compound, str) else compound
        compound_name = compound.Name if compound is not None else str(compound)
        if compound is None:
            raise ValueError(f"Compound '{compound_name}' not found in database.")
        return self.transfer_crits_from_compound(compound)

    def transfer_crits_from_compound(self, compound: Compound = None):
        """
        Set critical properties and Cp data from a Compound object.

        Parameters
        ----------
        compound: Compound
            Compound object containing critical properties and Cp data
        """
        if compound is not None:
            self.Tc = compound.Tc
            Pc_bar = compound.Pc
            if self.pressure_unit == 'MPa' or self.pressure_unit == 'mpa':
                self.Pc = Pc_bar / 10.0
            elif self.pressure_unit == 'bar':
                self.Pc = Pc_bar
            elif self.pressure_unit == 'kPa' or self.pressure_unit == 'kpa':
                self.Pc = Pc_bar * 100.0
            elif self.pressure_unit == 'Pa' or self.pressure_unit == 'pa':
                self.Pc = Pc_bar * 1.e5
            elif self.pressure_unit == 'atm':
                self.Pc = Pc_bar / 1.01325
            else:
                raise ValueError(f"Unsupported pressure unit: {self.pressure_unit}")
            self.omega = compound.Omega
            self.Cp = deepcopy(compound.Cp)
        return self


# Author: Cameron F. Abrams, <cfa22@drexel.edu>
from .eos import CubicEOS
from dataclasses import dataclass
import numpy as np
import logging

logger = logging.getLogger(__name__)

@dataclass
class IdealGasEOS(CubicEOS):
    """
    Ideal gas class derived from CubicEOS
    """

    description: str = "Ideal Gas Equation of State"

    def _calc_a(self):
        """ vdW a parameter for ideal gas is zero """
        return 0.0
    
    def _calc_b(self):
        """ vdW b parameter for ideal gas is zero """
        return 0.0
    
    @property
    def cubic_coeff(self) -> np.ndarray:
        """ cubic coefficients for ideal gas """
        return np.array([0.0, 0.0, 1.0, -1.0])

    @property
    def Z(self) -> np.ndarray:
        """ compressibility factor for ideal gas is unity """
        return np.array([1.0])
    
    def _calc_h_departure(self) -> np.ndarray:
        """ Enthalpy departure at state T and P; ideal-gas value """
        return np.array([0.0])

    def _calc_s_departure(self) -> np.ndarray:
        """ Entropy departure at state T and P; ideal-gas value """
        return np.array([0.0])
    
    def _calc_logphi(self) -> np.ndarray:
        """ natural log of fugacity coefficient at state T and P; ideal-gas value """
        return np.array([0.0])

    def solve(self, T: float = None, P: float = None, v: float = None):
        """
        Solve for missing property in ideal gas EOS

        Parameters
        ----------
        T : float, optional
            Temperature
        P : float, optional
            Pressure
        v : float, optional
            Molar volume
        """
        if T is not None:
            self.T = T
        if P is not None:
            self.P = P
        R = self.R_pv # volumetric gas constant
        if self.T is not None and self.P is not None:
            pass
        elif self.T is not None and v is not None:
            self.P = R * self.T / v
        elif self.P is not None and v is not None:
            self.T = self.P * v / R
        else:
            raise ValueError("Insufficient information to solve ideal gas EOS.")



from .idealgas import IdealGasEOS
from .vanderwaals import VanDerWaalsEOS
from .soaveredlichkwong import SoaveRedlichKwongEOS
from .pengrobinson import PengRobinsonEOS

__all__ = [
    "IdealGasEOS",
    "VanDerWaalsEOS",
    "SoaveRedlichKwongEOS",
    "PengRobinsonEOS",
]


# Author: Cameron F. Abrams, <cfa22@drexel.edu>

import os
import sys

import argparse as ap
import numpy as np

from importlib.metadata import version

from sandlerprops.properties import PropertiesDatabase
from sandlermisc.statereporter import StateReporter

from .eos import CubicEOS
from .idealgas import IdealGasEOS
from .soaveredlichkwong import SoaveRedlichKwongEOS
from .pengrobinson import PengRobinsonEOS
from .vanderwaals import VanDerWaalsEOS

banner = r"""
 __                 _ _           
/ _\ __ _ _ __   __| | | ___ _ __ 
\ \ / _` | '_ \ / _` | |/ _ \ '__|
_\ \ (_| | | | | (_| | |  __/ |   
\__/\__,_|_| |_|\__,_|_|\___|_|   
               _     _            
     ___ _   _| |__ (_) ___ ___   
    / __| | | | '_ \| |/ __/ __|  
   | (__| |_| | |_) | | (__\__ \  
    \___|\__,_|_.__/|_|\___|___/  v""" + version("sandlercubics") + """

"""

def bar_to_unit(value_in_bar: float, to_unit: str) -> float:
    """
    Convert pressure from bar to specified unit.
    
    Parameters
    ----------
    value_in_bar : float
        Pressure value in bar.
    to_unit : str
        Target unit ('mpa', 'MPa', 'kpa', 'kPa', 'pa', 'Pa', 'bar', 'atm').

    Returns
    -------
    float
        Converted pressure value.
    """
    conversions = {
        'mpa': 0.1,
        'MPa': 0.1,
        'kpa': 100.0,
        'kPa': 100.0,
        'pa': 1e5,
        'Pa': 1e5,
        'bar': 1.0,
        'atm': 0.986923,
    }
    if to_unit not in conversions:
        raise ValueError(f"Unsupported pressure unit conversion to {to_unit}.")
    return value_in_bar * conversions[to_unit]

def reporters(eos: CubicEOS) -> str:
    """
    Generate state and property reports for the given EOS.
    
    Parameters
    ----------
    eos : CubicEOS
        The equation of state object.

    Returns
    -------
    str
        Formatted state and property reports.
    """

    result = StateReporter({})
    
    result.add_property('T', eos.T, 'K', fstring="{: .2f}")
    pu = eos.R._capitalizations.get(eos.pressure_unit, eos.pressure_unit)
    result.add_property('P', eos.P, pu, fstring="{: .2f}")
    result.add_property('Z', ', '.join([f"{z: 4g}" for z in eos.Z]), '', fstring=None)
    vu = eos.R._capitalizations.get(eos.volume_unit, eos.volume_unit)
    result.add_property('v', ', '.join([f"{vol: 6g}" for vol in eos.v]), f'{vu}/mol', fstring=None)
    result.add_property('h', ', '.join([f"{h: 6g}" for h in eos.h]), 'J/mol', fstring=None)
    result.add_property('s', ', '.join([f"{s: 6g}" for s in eos.s]), 'J/mol-K', fstring=None)
    result.add_property('hdep', ', '.join([f"{hdep: 6g}" for hdep in eos.h_departure]), 'J/mol', fstring=None)
    result.add_property('sdep', ', '.join([f"{sdep: 6g}" for sdep in eos.s_departure]), 'J/mol-K', fstring=None)
    if eos.T < eos.Tc:
        pu = eos.R._capitalizations.get(eos.pressure_unit, eos.pressure_unit)
        if eos.Pvap is not np.nan:
            result.add_property(f'Pvap({eos.T:.2f} K)', eos.Pvap, pu, fstring="{:.2f}")
        if eos.Hvap is not np.nan:
            result.add_property(f'Hvap({eos.T:.2f} K)', eos.Hvap, 'J/mol', fstring="{:.2f}")
        if eos.Svap is not np.nan:
            result.add_property(f'Svap({eos.T:.2f} K)', eos.Svap, 'J/mol-K', fstring="{:.4f}")
        if eos.P < eos.Pc:
            pu = eos.R._capitalizations.get(eos.pressure_unit, eos.pressure_unit)
            if eos.Tsat is not np.nan:
                result.add_property(f'Tsat({eos.P:.2f} {pu})', eos.Tsat, 'K', fstring="{:.2f}")
        if eos.x is not None:
            result.add_property('Vapor fraction x', eos.x, '', fstring="{:4g}")
    prop = StateReporter({})
    prop.add_property('Tc', eos.Tc, 'K', fstring="{:.2f}")
    pu = eos.R._capitalizations.get(eos.pressure_unit, eos.pressure_unit)
    prop.add_property('Pc', eos.Pc, pu, fstring="{:.2f}")
    prop.add_property('omega', eos.omega, '', fstring="{:.3f}")

    prop.add_property('Tref', eos.Tref, 'K', fstring="{:.2f}")
    prop.add_property('Pref', eos.Pref_MPa, 'MPa', fstring="{:.2f}")
    # print(eos.Cp)
    prop.pack_Cp(eos.Cp, fmts=["{:.2f}", "{:.3e}", "{:.3e}", "{:.3e}"])
    return result.report(),  prop.report()

def state(args):
    """
    Calculate and report the state for a single condition using the specified EOS.

    Parameters
    ----------
    args : argparse.Namespace
        Parsed command-line arguments.
    """
    pressure_unit = args.pressure_unit
    volume_unit = args.volume_unit
    if args.n is not None:
        db = PropertiesDatabase()
        component = db.get_compound(args.n)
        if component is None:
            print(f"Component '{args.n}' not found in database.")
            return
    match args.eos_type:
        case 'ideal':
            eos = IdealGasEOS(pressure_unit=pressure_unit, volume_unit=volume_unit)
        case 'vdw':
            eos = VanDerWaalsEOS(pressure_unit=pressure_unit, volume_unit=volume_unit)
        case 'pr':
            eos = PengRobinsonEOS(pressure_unit=pressure_unit, volume_unit=volume_unit)
        case 'srk':
            eos = SoaveRedlichKwongEOS(pressure_unit=pressure_unit, volume_unit=volume_unit)
    if component is not None:
        eos.set_compound(component)
    if args.Tc is not None:
        eos.Tc = args.Tc
    if args.Pc is not None:
        eos.Pc = args.Pc
    if args.w is not None:
        eos.omega = args.w
    if args.Cp is not None:
        eos.Cp = args.Cp
    if args.eos_type == 'ideal':
        eos.solve(T=args.T, P=args.P, v=args.v)
    else:
        eos.solve(T=args.T, P=args.P, v=args.v, h=args.h, s=args.s, u=args.u)
    state_report, prop_report = reporters(eos)
    print(f'State report for {component.Name} using {eos.description}:')
    print(state_report)
    if prop_report and args.show_props:
        print("\nConstants used for calculations:")
        print(prop_report)

def delta(args):
    """
    Calculate and report property differences between two states using the specified EOS.

    Parameters
    ----------
    args : argparse.Namespace
        Parsed command-line arguments.
    """
    pressure_unit = args.pressure_unit
    volume_unit = args.volume_unit
    if args.n is not None:
        db = PropertiesDatabase()
        component = db.get_compound(args.n) # critical pressures are in bars!
        if component is None:
            print(f"Component '{args.n}' not found in database.")
            return
    match args.eos_type:
        case 'ideal':
            eos1 = IdealGasEOS(pressure_unit=pressure_unit, volume_unit=volume_unit)
            eos2 = IdealGasEOS(pressure_unit=pressure_unit, volume_unit=volume_unit)
        case 'vdw':
            eos1 = VanDerWaalsEOS(pressure_unit=pressure_unit, volume_unit=volume_unit)
            eos2 = VanDerWaalsEOS(pressure_unit=pressure_unit, volume_unit=volume_unit)
        case 'pr':
            eos1 = PengRobinsonEOS(pressure_unit=pressure_unit, volume_unit=volume_unit)
            eos2 = PengRobinsonEOS(pressure_unit=pressure_unit, volume_unit=volume_unit)
        case 'srk':
            eos1 = SoaveRedlichKwongEOS(pressure_unit=pressure_unit, volume_unit=volume_unit)
            eos2 = SoaveRedlichKwongEOS(pressure_unit=pressure_unit, volume_unit=volume_unit)
    if component is not None:
        eos1.set_compound(component)
        eos2.set_compound(component)
    if args.Tc is not None:
        eos1.Tc = args.Tc
        eos2.Tc = args.Tc
    if args.Pc is not None:
        eos1.Pc = args.Pc
        eos2.Pc = args.Pc
    if args.w is not None:
        eos1.omega = args.w
        eos2.omega = args.w
    if args.Cp is not None:
        eos1.Cp = args.Cp
        eos2.Cp = args.Cp
    if args.eos_type == 'ideal':
        eos1.solve(T=args.T1, P=args.P1, v=args.v1)
        eos2.solve(T=args.T2, P=args.P2, v=args.v2)
    else:
        eos1.solve(T=args.T1, P=args.P1, v=args.v1, h=args.h1, s=args.s1, u=args.u1)
        eos2.solve(T=args.T2, P=args.P2, v=args.v2, h=args.h2, s=args.s2, u=args.u2)

    delta_State = StateReporter({})
    delta_H = eos2.h - eos1.h
    delta_S = eos2.s - eos1.s
    delta_U = eos2.u - eos1.u
    delta_State.add_property('Δh', ', '.join(f'{x: 7g}' for x in delta_H), 'J/mol', fstring=None)
    delta_State.add_property('Δs', ', '.join(f'{x: 7g}' for x in delta_S), 'J/mol-K', fstring=None)
    delta_State.add_property('Δu', ', '.join(f'{x: 7g}' for x in delta_U), 'J/mol', fstring=None)
    state_1, _ = reporters(eos1)
    state_2, consts = reporters(eos2)
    print(f"State-change calculations for {component.Name} using {eos1.description}:")
    if args.show_states:
        print()
        two_states = ["State 1:                       State 2:"]
        for line1, line2 in zip(state_1.splitlines(), state_2.splitlines()):
            two_states.append(f"{line1:<26s}     {line2}")
        print("\n".join(two_states))
        print()
        print("Property changes:")
    print(delta_State.report())
    if args.show_props:
        print("\nConstants used for calculations:")
        print(consts)
    
def cli():
    """
    Command-line interface for sandlercubics package.
    """

    subcommands = {
        'state': dict(
            func = state,
            help = 'work with a cubic equation of state for a single state'
        ),
        'delta': dict(
            func = delta,
            help = 'work with property differences between two states (not implemented yet)'
        ),
    }
    parser = ap.ArgumentParser(
        prog='sandlercubics',
        description="Sandlercubics: A collection of computational tools using cubic equations of state based on Chemical, Biochemical, and Engineering Thermodynamics (5th edition) by Stan Sandler",
        epilog="(c) 2025, Cameron F. Abrams <cfa22@drexel.edu>"
    )
    parser.add_argument(
        '-b',
        '--banner',
        default=False,
        action=ap.BooleanOptionalAction,
        help='toggle banner message'
    )
    parser.add_argument(
        '-v',
        '--version',
        action='version',
        version=f'sandlercubics version {version("sandlercubics")}',
        help='show program version and exit'
    )
    subparsers = parser.add_subparsers(
        title="subcommands",
        dest="command",
        metavar="<command>",
        required=True,
    )
    command_parsers={}
    for k, specs in subcommands.items():
        command_parsers[k] = subparsers.add_parser(
            k,
            help=specs['help'],
            add_help=False,
            formatter_class=ap.RawDescriptionHelpFormatter
        )
        command_parsers[k].set_defaults(func=specs['func'])
        command_parsers[k].add_argument(
            '--help',
            action='help',
            help=specs['help']
        )

    options = [
        ('eos', 'eos_type', 'type of cubic equation of state to use', str, 'vdw', ['ideal', 'vdw', 'pr', 'srk']),
        ('pu', 'pressure_unit', 'pressure unit (mpa, bar, atm)', str, 'mpa', ['mpa', 'bar', 'atm']),
        ('vu', 'volume_unit', 'volume unit (m3, l, cm3)', str, 'm3', ['m3', 'l', 'cm3']),
    ]
    for short, long, desc, typ, default, choices in options:
        command_parsers['state'].add_argument(
            f'-{short}',
            f'--{long}',
            dest=long,
            type=typ,
            default=default,
            choices=choices,
            help=desc
        )
        command_parsers['delta'].add_argument(
            f'-{short}',
            f'--{long}',
            dest=long,
            type=typ,
            default=default,
            choices=choices,
            help=desc
        )

    crit_args = [
        ('n', 'component', 'component name (e.g., methane, ethane, etc.)', str, False),
        ('Pc', 'critical_pressure', 'critical pressure (if component not specified)', float, False),
        ('Tc', 'critical_temperature', 'critical temperature in K (if component not specified)', float, False),
        ('w', 'acentric_factor', 'acentric factor omega (if component not specified)', float, False),
    ]

    cp_args = [
        ('CpA', 'CpA', 'heat capacity coefficient A (J/mol-K) (if component not specified)', float, False),
        ('CpB', 'CpB', 'heat capacity coefficient B (J/mol-K^2) (if component not specified)', float, False),
        ('CpC', 'CpC', 'heat capacity coefficient C (J/mol-K^3) (if component not specified)', float, False),
        ('CpD', 'CpD', 'heat capacity coefficient D (J/mol-K^4) (if component not specified)', float, False),
    ]

    state_args = [
        ('P', 'pressure', 'pressure', float, False),
        ('T', 'temperature', 'temperature in K (always in K)', float, False),
        ('v', 'molar_volume', 'molar volume', float, False),
        ('h', 'enthalpy', 'molar enthalpy in J/mol', float, False),
        ('s', 'entropy', 'molar entropy in J/mol-K', float, False),
        ('u', 'internal_energy', 'molar internal energy in J/mol', float, False),
    ]
    for prop, long_arg, explanation, arg_type, required in state_args + crit_args:
        command_parsers['state'].add_argument(
            f'-{prop}',
            f'--{long_arg}',
            dest=prop,
            type=arg_type,
            required=required,
            help=explanation
        )
    
    command_parsers['state'].add_argument(
        '--Cp',
        nargs=4,
        type=float,
        metavar=('CpA', 'CpB', 'CpC', 'CpD'),
        help='heat capacity polynomial coefficients A, B, C, D (J/mol-K, J/mol-K^2, J/mol-K^3, J/mol-K^4) (if component not specified)',
        default=None
    )
    command_parsers['state'].add_argument(
        '--show-props',
        default=False,
        action=ap.BooleanOptionalAction,
        help='also show all critical properties and Cp coefficients used'
    )

    delta_args = [
        ('P1', 'pressure1', 'pressure of state 1', float, False),
        ('T1', 'temperature1', 'temperature of state 1 in K (always in K)', float, False),
        ('v1', 'molar_volume1', 'molar volume of state 1', float, False),
        ('h1', 'enthalpy1', 'molar enthalpy of state 1 in J/mol', float, False),
        ('s1', 'entropy1', 'molar entropy of state 1 in J/mol-K', float, False),
        ('u1', 'internal_energy1', 'molar internal energy of state 1 in J/mol', float, False),
        ('P2', 'pressure2', 'pressure of state 2', float, False),
        ('T2', 'temperature2', 'temperature of state 2 in K (always in K)', float, False),
        ('v2', 'molar_volume2', 'molar volume of state 2', float, False),
        ('h2', 'enthalpy2', 'molar enthalpy of state 2 in J/mol', float, False),
        ('s2', 'entropy2', 'molar entropy of state 2 in J/mol-K', float, False),
        ('u2', 'internal_energy2', 'molar internal energy of state 2 in J/mol', float, False),
    ]
    for prop, long_arg, explanation, arg_type, required in delta_args + crit_args:
        command_parsers['delta'].add_argument(
            f'-{prop}',
            f'--{long_arg}',
            dest=prop,
            type=arg_type,
            required=required,
            help=explanation
        )

    command_parsers['delta'].add_argument(
        '--show-states',
        default=False,
        action=ap.BooleanOptionalAction,
        help='also show the full states for state 1 and state 2'
    )
    command_parsers['delta'].add_argument(
        '--show-props',
        default=False,
        action=ap.BooleanOptionalAction,
        help='also show all critical properties and Cp coefficients used'
    )
    command_parsers['delta'].add_argument(
        '--Cp',
        nargs=4,
        type=float,
        metavar=('CpA', 'CpB', 'CpC', 'CpD'),
        help='heat capacity polynomial coefficients A, B, C, D (J/mol-K, J/mol-K^2, J/mol-K^3, J/mol-K^4) (if component not specified)',
        default=None
    )

    args = parser.parse_args()

    if args.func == state:
        nprops = 0
        for prop, _, _, _, _ in state_args:
            if hasattr(args, prop) and getattr(args, prop) is not None:
                nprops += 1
        if nprops > 2:
            parser.error('At most two of P, T, x, v, u, h, and s may be specified for "state" subcommand')

    if args.banner:
        print(banner)
    if hasattr(args, 'func'):
        args.func(args)
    else:
        my_list = ', '.join(list(subcommands.keys()))
        print(f'No subcommand found. Expected one of {my_list}')
    if args.banner:
        print('Thanks for using sandlercubics!')


# Author: Cameron F. Abrams, <cfa22@drexel.edu>
from .eos import CubicEOS
from dataclasses import dataclass
import numpy as np
import logging

logger = logging.getLogger(__name__)

@dataclass
class VanDerWaalsEOS(CubicEOS):
    """
    Generalized van der Waals equation of state, in which 'a' and 'b'
    are calculated from critical constants
    """

    description: str = "Van der Waals Equation of State"

    def _calc_a(self):
        """
        Calculates parameter a for Van der Waals EOS
        """
        return (27 / 64) * self.R**2 * self.Tc**2 / self.Pc
    
    def _calc_b(self):
        """
        Calculates parameter b for Van der Waals EOS
        """
        return self.R * self.Tc / (8 * self.Pc)

    @property
    def cubic_coeff(self):
        """
        cubic coefficients for Van der Waals EOS
        """
        A, B = self.A, self.B
        return np.array([1.0, -1.0 - B, A, -A * B])

    def _calc_h_departure(self) -> np.ndarray:
        """
        Enthalpy departure for Van der Waals EOS
        """
        return self.R * self.T * (self.Z - 1 - self.A * np.reciprocal(self.Z))

    def _calc_s_departure(self) -> np.ndarray:
        """
        Entropy departure for Van der Waals EOS
        """
        return self.R * (np.log(self.Z - self.B) - self.A / self.Z)

    def _calc_logphi(self) -> np.ndarray:
        """
        natural log of fugacity coefficient for Van der Waals EOS
        """
        return self.Z - 1 - np.log(self.Z - self.B) - self.A / self.Z


# Author: Cameron F. Abrams, <cfa22@drexel.edu>

from .eos import CubicEOS, sqrt_2
from dataclasses import dataclass
import numpy as np
import logging

logger = logging.getLogger(__name__)

@dataclass
class SoaveRedlichKwongEOS(CubicEOS):
    """
    Pure-component Soave-Redlich-Kwong equation of state (Eq. 6.4-1 in Sandler 5th ed)
    """

    description: str = "Soave-Redlich-Kwong Equation of State"

    @property
    def kappa(self):
        """
        kappa parameter for Soave-Redlich-Kwong EOS
        """
        return 0.480 + 1.574 * self.omega - 0.176 * self.omega**2
    
    @property
    def alpha(self):
        """
        alpha parameter for Soave-Redlich-Kwong EOS
        """
        return (1 + self.kappa * (1 - np.sqrt(self.T/self.Tc)))**2

    def _calc_a(self):
        """
        Calculates parameter a for Soave-Redlich-Kwong EOS
        """
        return 0.42748 * self.R**2 * self.Tc**2 / self.Pc * self.alpha
    
    def _calc_b(self):
        """
        Calculates parameter b for Soave-Redlich-Kwong EOS
        """
        return 0.08664 * self.R * self.Tc / self.Pc

    @property
    def cubic_coeff(self):
        """
        cubic coefficients for Soave-Redlich-Kwong EOS
        """
        A, B = self.A, self.B
        return np.array([1.0, -1.0, A - B - B**2, -A * B])

    @property
    def lrfrac(self) -> np.ndarray:
        """
        helper function for ln fugacity, enthalpy, and entropy departures
        """
        z = self.Z
        num_arg = z + self.B
        den_arg = z
        return np.log(num_arg / den_arg)
    
    def _calc_h_departure(self) -> np.ndarray:
        """
        Enthalpy departure at state T and P (from solution to problem 6.36 in Sandler 5th ed)
        """
        z = self.Z
        return self.R * self.T * (z - 1) + (self.T * self.da_dT - self.a) / self.b * self.lrfrac

    def _calc_s_departure(self) -> np.ndarray:
        """
        Entropy departure at state T and P (from solution to problem 6.36 in Sandler 5th ed)
        """
        z = self.Z
        return self.R * np.log(z - self.B) + self.da_dT/self.b * self.lrfrac

    def _calc_logphi(self) -> np.ndarray:
        """
        natural log of fugacity coefficient at state T and P (from solution to problem 6.46 in Sandler 5th ed)
        """
        z = self.Z
        return z - 1 - np.log(z - self.B) - self.a / (self.R * self.T * self.b) * self.lrfrac